############################################################################
# booit.pl -- (c) 2002 Y. Aulchenko, EMC Rotterdam, Inst. Cytol. Genet. SD RAS
# contact: yuriisa@bionet.nsc.ru
# This program is distributed under the terms of GNU Public license (GPL)
#
# Interface over likLD(lsqLD)/booitf -- generates bootstrap-samples 
# from d' data, call estimation procedure (likLD/lsqLD), records par. 
# estimates, use distribution to get CIs and var/cov matrix.
#
# created -- 04.09.2002
# modified -- 05.09.2002
#
############################################################################

############################################################################
# DOS. Uncomment the following section for DOS platform and comment for UNIX
#$sboo="./booitf.exe";
############################################################################
# UNIX. Uncomment the following section for Unix-like platform and comment for DOS
$sboo="./booitf";

print "BooIt ver. 1.1 -- bootstrap distribution for parameter estimates\n";

print "
Usage: perl booit.pl -p program -d data -n number -o output

\twhere program is likLD or lsqLD
\tdata is data-file (d-primes vs. distance)
\tnumber is number of bootstraps to perform
\toutput is output file name\n
";

$prog = "./likLD";
$data = "dprim.dat";
$nboos = 1000;
$outf = "boores.txt";

$line = join (' ',@ARGV);
$prog=$1 if ($line =~ /-p *(.[^\s]*)/);
$prog="./$prog" if ($prog !~ /^\.\//);
$data=$1 if ($line =~ /-d *(.[^\s]*)/);
$nboos=$1 if ($line =~ /-n *(.[^\s]*)/);
$outf=$1 if ($line =~ /-o *(.[^\s]*)/);

print "Options in effect:
\tprogname : $prog
\tdata     : $data
\tnumber   : $nboos
\toutput   : $outf

";

$time=localtime();
$time0 = time();
print"Run started $time\n\n";

die "wrong program name (use likLD or lsqLD)\n" if ($prog ne "./likLD" and $prog ne "./lsqLD");
die "wrong number of bootstraps (use positive integer)\n" if ($nboos =~ /[^0-9]/ or $nboos<=0);

# copy data-file to in.boo (used by booitf (fortran bootstrap-prog))
open (TMP,"$data") or die "cannot open data-file $data: $!\n";
open (TMP1,">in.boo") or die "cannot open in.boo for writing: $!\n";
while (<TMP>) {print TMP1 $_;} 
close TMP1;
close TMP;

open (OUT,">booit.dump") or die "cannot open boo-dump-file booit.dump for output: $!\n";

$fstrun=0;
for ($i=0;$i<$nboos;$i++) {
# boo the file in.boo (output is in out.boo)
	`$sboo`;
# do estimation with boo'ed file
	`$prog out.boo > boores.boo.tmp`;
# process results
	open(BOOF,"boores.boo.tmp");@boores = <BOOF>;close BOOF;
# process hypothesis 3
	$line = $boores[12];
	@line = split /\s+/,$line;
	$h3p1[$i] = $line[4];
	$line = $boores[13];
	@line = split /\s+/,$line;
	$h3p2[$i] = $line[4];
	$line = $boores[14];
	@line = split /\s+/,$line;
	$h3p3[$i] = $line[4];
	$line = $boores[15];
	@line = split /\s+/,$line;
	$h3p4[$i] = $line[4];

# process hypothesis 2
	$line = $boores[39];
	@line = split /\s+/,$line;
	$h2p1[$i] = $line[4];
	$line = $boores[40];
	@line = split /\s+/,$line;
	$h2p2[$i] = $line[4];
	$line = $boores[41];
	@line = split /\s+/,$line;
	$h2p3[$i] = $line[4];
	$line = $boores[42];
	@line = split /\s+/,$line;
	$h2p4[$i] = $line[4];

# process hypothesis 1
	$line = $boores[65];
	@line = split /\s+/,$line;
	$h1p1[$i] = $line[4];
	$line = $boores[66];
	@line = split /\s+/,$line;
	$h1p2[$i] = $line[4];
	$line = $boores[67];
	@line = split /\s+/,$line;
	$h1p3[$i] = $line[4];
	$line = $boores[68];
	@line = split /\s+/,$line;
	$h1p4[$i] = $line[4];

# process hypothesis 0
	$line = $boores[90];
	@line = split /\s+/,$line;
	$h0p1[$i] = $line[4];
	$line = $boores[91];
	@line = split /\s+/,$line;
	$h0p2[$i] = $line[4];
	$line = $boores[92];
	@line = split /\s+/,$line;
	$h0p3[$i] = $line[4];
	$line = $boores[93];
	@line = split /\s+/,$line;
	$h0p4[$i] = $line[4];
	$nunu=$i+1;
	print OUT "boo=$nunu $h3p1[$i] $h3p2[$i] $h3p3[$i] $h3p4[$i] $h2p1[$i] $h2p2[$i] $h2p3[$i] $h2p4[$i] $h1p1[$i] $h1p2[$i] $h1p3[$i] $h1p4[$i] $h0p1[$i] $h0p2[$i] $h0p3[$i] $h0p4[$i] \n";

	$fstrun++;
	if ($fstrun == 1) {
		$took = time() - $time0;
		$enum=$nboos;
		$etime = $took * $enum;
		print "First run took $took seconds. Estimated running time is $etime seconds\n";
		print "Estimated time is very innacurate (1st run-time <3 sec.)\n" if ($took<3);
		print "\n";
	}
	print "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\bbooing # $nunu";
}

close OUT;
print "\n\n";

# number of quantile-elements
$q1 = int ($nboos*0.025+1.0);
$q2 = int ($nboos*0.975);

open (OUT,">$outf") or die "cannot open output-file $outf for output: $!\n";

print OUT "Quantile-elements used: 2.5\% -- $q1, 97.5\% -- $q2\n";

for ($i=0;$i<4;$i++) {
	for ($j=1;$j<=4;$j++) {
		$name = "h${i}p${j}"; 
		@arr1 = @{$name};
		for ($j1=$j;$j1<=4;$j1++) {
			$name = "h${i}p${j1}"; 
			@arr2 = @{$name};
			$covxx = &covar();
			if ($j == $j1) {
				@so = sort {$a <=> $b} @{$name};
				$left = @so[$q1];$right=$so[$q2];
				print OUT "Bootstrap s.e. (2.5\% -- 97.5\%) for H$i parameter $j: $left -- $right\n";
			}
			print OUT "H$i,COV($j,$j1) = $covxx (CORR($j,$j1) = $corrxy)\n";
		}
	}
}

$time=localtime();
$timeN=time();
$took = $timeN-$time0;
print"\n\nRun finished $time in $took seconds\n\n";

print "\nResults written to $outf\nComplete dump of bootstrapped parameters is in booit.dump\n\n";


sub covar {
	my $sumx=0.,$sumy=0.,$sumprod=0.,$mex,$mey,$imi,$cov,$ssqx=0.,$ssqy=0.;
	for ($imi=0;$imi<@arr1;$imi++) {
		$sumx+=$arr1[$imi];
		$ssqx+=$arr1[$imi]*$arr1[$imi];
		$sumy+=$arr2[$imi];
		$ssqy+=$arr2[$imi]*$arr2[$imi];
		$sumprod+=$arr1[$imi]*$arr2[$imi];
	}
	my $N = (@arr1);
	my $Nm1 = $N -1.0;
	$mex = $sumx / $N;
	$mey = $sumy / $N;
	$cov = ($sumprod - $N*$mex*$mey)/$Nm1;
	$varx = ($ssqx - $N*$mex*$mex)/$Nm1;
	$vary = ($ssqy - $N*$mey*$mey)/$Nm1;
	if ($vary>0. and $varx>0.) {
		$corrxy = $cov/(sqrt($vary)*sqrt($varx));
	} else {
		$corrxy="NA";
	}
	return $cov;
}
