#include <methgi.h>
#include <math.h>
//############################################################################
//# lsqLD -- (c) 2002 Y. Aulchenko, EMC Rotterdam, Inst. Cytol. Genet. SD RAS
//# contact: yuriisa@bionet.nsc.ru
//# This program is distributed under the terms of GNU Public license (GPL)
//#
//# the program is just a copy of likLD.cpp -- but normal distribution 
//# was eliminated. Now it does <weighted> least squares. The 4th parameter
//# (s2e) does not make any sense now. Output max.-function-value is negative
//# of sum of squares. Together with # observations and number of paramaters
//# under estimation (no variance anymore!) this could be used for hypotheses
//# testing 
//#
//# created -- 05.09.2002
//# last modified -- 06.09.2002
//############################################################################

double fun();

# define MAXLEN 20000

double 	theta[MAXLEN];
double 	dprim[MAXLEN];
double 	weight[MAXLEN];
int	numlin;

double fun (double *p, int size) {
//
//    
// Current values of parameters are stored in "p"
// Read these values.
//
	double M,L,T,s2e;
	M=p[0];
	L=p[1];
	T=p[2];
	s2e=p[3]*p[3];
//
//  
// Check additional constains
//

//	if (M<L) return NEGLI_LH; 	// if the constain is not fulfilled,
					// return negligable value 

//
//  
// Calculation of likelihood function
//
	double LH=0.;
	double roexp;
	for (int i=0;i<numlin;i++) {
		roexp 	=  L + (M-L)*exp(-theta[i]*T);
		LH 	+= -weight[i]*(roexp-dprim[i])*(roexp-dprim[i]);
	}
//	cout << M << " " << L << " " << T << " " << s2e << " " << LH << " " << "\n";
	return LH;
}

int load_data(char *fname) {
//
//    
// Loading data from file
//
	FILE *dafa;
	char string[256];
	if (!(dafa=fopen(fname,"r"))) return 1;
	if (!fgets(string,256,dafa)) return 2;
	if (!fgets(string,256,dafa)) return 2;
	if (!sscanf(string,"%d",&numlin)) return 3;
	if (numlin > MAXLEN) {
		cout << "MAXLEN exceeded! change and recompile.\n";
		return 2;
	}
	for (int i=0;i<numlin;i++) {
		if (!fgets(string,256,dafa)) return 2;
		if (!sscanf(string,"%lf %lf %lf",&theta[i],&dprim[i],&weight[i])) return 3;
//		cout << i << " " << num[i] << " " << theta[i] << "\n";
	}
	fclose(dafa);
	return 0;
}

main(int narg, char **arg) {
	if (narg<2) {
		cout << "Usage: likLD infile\n";
		return 1;
	}
	char *datafile;
	datafile = arg[1];
//	char datafile[]="dprim.dat";
	if ((load_data(datafile))) {
		cout << "     \n/cannot loda data from file/ " << datafile << endl;
		return 1;
	}
//
//   par  methgi,    
//    par.inp
// Create object par of the class methgi, initial values of parameters
// are in file par.inp
//
//	cout << "------------------------------------------------------------" << endl;
//	cout << "   /initial values of parameters/\n";
//
//   (.. )  
// Print out current (i.e. initial) values of parameters, etc.
//
//	par.printout();
//
//       
// Maximisation of function and printout the results
//
	cout<<"-----------------MAXIMISING UNDER H3 -----------------\n";
	methgi par2("par3.inp");
	cout << par2.maximise("out3.par",fun,1) << endl;
	cout<<"maximum likelihood = "<<par2.getLH()<<"\n";
	cout<<"model: Dprim = "<<par2.value(1)<<" + ("<<par2.value(0)<<" - "<<par2.value(1)<<")*Exp(-Theta*"<<par2.value(2)<<")\n";
	cout<<"s2res = "<<par2.value(3)*par2.value(3)<<"\n95% CI for parameters:\n";
	for (int i=0;i<par2.npar();i++) {
//	cout<<"par "<<i<<" : "<<par2.left95(fun,i)<<" - "<<par2.right95(fun,i)<<"\n";	
	cout<<"par "<<i<<" : "<<"      NA        "<<" - "<<"      NA         "<<"\n";	
	}
	cout<<"------------ FINISHED MAXIMISING UNDER H3 ------------\n";
	cout<<"-----------------MAXIMISING UNDER H2 -----------------\n";
	methgi par1("par2.inp");
	cout << par1.maximise("out2.par",fun,1) << endl;
	cout<<"maximum likelihood = "<<par1.getLH()<<"\n";
	cout<<"model: Dprim = "<<par1.value(1)<<" + ("<<par1.value(0)<<" - "<<par1.value(1)<<")*Exp(-Theta*"<<par1.value(2)<<")\n";
	cout<<"s2res = "<<par1.value(3)*par1.value(3)<<"\n95% CI for parameters:\n";
	for (int i=0;i<par1.npar();i++) {
	if (i!=1) {
//	cout<<"par "<<i<<" : "<<par1.left95(fun,i)<<" - "<<par1.right95(fun,i)<<"\n";	
	cout<<"par "<<i<<" : "<<"      NA        "<<" - "<<"      NA         "<<"\n";	
	}
	}
	cout<<"------------ FINISHED MAXIMISING UNDER H2 ------------\n";
	cout<<"-----------------MAXIMISING UNDER H1 -----------------\n";
	methgi par10("par1.inp");
	cout << par10.maximise("out1.par",fun,1) << endl;
	cout<<"maximum likelihood = "<<par10.getLH()<<"\n";
	cout<<"model: Dprim = "<<par10.value(1)<<" + ("<<par10.value(0)<<" - "<<par10.value(1)<<")*Exp(-Theta*"<<par10.value(2)<<")\n";
	cout<<"s2res = "<<par10.value(3)*par10.value(3)<<"\n95% CI for parameters:\n";
	for (int i=0;i<par10.npar();i++) {
	if (i!=0 && i!=1) {
//	cout<<"par "<<i<<" : "<<par10.left95(fun,i)<<" - "<<par10.right95(fun,i)<<"\n";	
	cout<<"par "<<i<<" : "<<"      NA        "<<" - "<<"      NA         "<<"\n";	
	}
	}
	cout<<"------------ FINISHED MAXIMISING UNDER H1 ------------\n";
	cout<<"-----------------MAXIMISING UNDER H0 -----------------\n";
	methgi par("par0.inp");
	cout << par.maximise("out0.par",fun,1) << endl;
	cout<<"maximum likelihood = "<<par.getLH()<<"\n";
	cout<<"model: Dprim = "<<par.value(1)<<" + ("<<par.value(0)<<" - "<<par.value(1)<<")*Exp(-Theta*"<<par.value(2)<<")\n";
	cout<<"s2res = "<<par.value(3)*par.value(3)<<"\n95% CI for parameters:\n";
	for (int i=0;i<par.npar();i++) {
	if (i!=1 && i!=2) {
//	cout<<"par "<<i<<" : "<<par.left95(fun,i)<<" - "<<par.right95(fun,i)<<"\n";	
	cout<<"par "<<i<<" : "<<"      NA        "<<" - "<<"      NA         "<<"\n";	
	}
	}
	cout<<"------------ FINISHED MAXIMISING UNDER H0 ------------\n";

        return 0;
}

