#include <stdio.h>
#include <iostream>
#include <fstream>
using namespace std;

//############################################################################
//# METHGI -- (c) Fedotov, Ginzburg, Axenovich, Aulchenko
//# This program is distributed under the terms of GNU Public license (GPL)
//############################################################################

#define MAXNPAR 20
#define NEGLI_LH -1.e+18
#define PER 1.e-4
#define NI 500
#define RED 0.5
#define EPS 1.e-10
#define CHI2 = 3.841455

typedef double (*LHfunction) (double *par, int size);

double list[MAXNPAR];

void ABTMsg(char *msg);
void ABTMsg(char *msg1, char *msg2);
void ABTMsg(char *msg1, int msg2);
double tiabs(double v);

class methgi {
	int 	NITER;
	int 	NFUN;
	double	H;
        int     NPAR;
	char*	NAME[MAXNPAR];
	int 	IVEC[MAXNPAR];
	int	ICP[MAXNPAR];
        double  PR[MAXNPAR];
        double  ST[MAXNPAR];
	double 	GL[MAXNPAR];
	double 	GR[MAXNPAR];
public:
        methgi(char *fname);
	int npar();
        double value(int pnum);
        char* name(int pnum);
	void write(char *fnout,int ifmax);
	double *parlist();
	double left95(LHfunction FUN,int numpar);
	double right95(LHfunction FUN,int numpar);
	double getLH();
	char* maximise(char *fnam, LHfunction fun, int outmode);
	int checkpar();
	void printout(double curlh);
	void printout();
};

methgi::methgi(char *fname) {
//
//     fname
//
	FILE *ipf; 
        cout << "    /loading parameters from file/ " << fname << endl;
	ipf=fopen(fname,"r");
	if (!ipf) ABTMsg("    /cannot open file/ ",fname);
	char string[256]; 
	if (!fgets(string,256,ipf)) ABTMsg("   /corrupted parameter file/ ",fname);
	if (!sscanf(string,"%d",&NPAR)) ABTMsg("   /corrupted parameter file/ ",fname);
	for (int i=0;i<NPAR;i++) {
		if (!fgets(string,256,ipf)) ABTMsg("   /corrupted parameter file/ ",fname);
		if (!sscanf(string,"%d %d %lf %lf %lf %lf",&IVEC[i],&ICP[i],&PR[i],&ST[i],&GL[i],&GR[i])) ABTMsg("   /corrupted parameter file/ ",fname);	
	}
	fclose(ipf);
//
//  . 
//   ,   
//
	for (int i=0;i<NPAR;i++) {
	  if (IVEC[i]>(i+1)) ABTMsg("IVEC       , \nIVEC should be less or equal to the number of parameter, parameter ... ",i);
	  if (IVEC<=0) ABTMsg("IVEC    0,  /IVEC should be > 0, parameter/ ",i);
	  if (IVEC[i]!=(i+1)) {
	    cout << "  /equality of parameters/ "<<(i+1)<< " and " << IVEC[i] << endl;
	    PR[i]=PR[IVEC[i]-1];
            ICP[i]=0;
 	  }
	}
	for (int i=0;i<NPAR;i++) {
		if (GL[i]>=GR[i]) ABTMsg("   \n/incorrect boundaries for parameter/ ... ",i+1);
		if (ST[i]>=(GR[i]-GL[i])) ABTMsg("    \n/initial step of the parameter is too big/ ... ",i+1);
		if (PR[i]>GR[i] || PR[i]<GL[i]) ABTMsg("     \n/initial parameter's value is out of the boundaries/ ... ",i+1);
		if (ICP[i]!=0 && (tiabs(PR[i]-GR[i])<1.e-6 || tiabs(PR[i]-GL[i])<1.e-6)) ABTMsg(" ,      \n/parameter is under maximisation, initial parameter's value should be INSIDE of the boundaries/ ... ",i+1);
	}
//
//   ,  
//
	cout << "  /number of parameters/ = " << NPAR << endl;
	cout << "     \n/succesful reading and verification of parameters/" << endl;
}

char * methgi::maximise(char *fnam, LHfunction FUN, int outmode) {
	double temppar[MAXNPAR];
	int tempnpar;
	tempnpar=NPAR;
	double Y[MAXNPAR],X,Z;
	double eps=1.e-17,BEGIN,F,F1,FF;
	int KK,N1,N,J,K,KP;
	double R=RED;
	double EP=EPS;
	N=NPAR;
i1:     KK=0;
        NITER=0;
        NFUN=1;
        BEGIN=NEGLI_LH;
        N1=0;
	for (J=0;J<N;J++) {
        	N1=N1+ICP[J];
	       	Y[J]=PR[J]; 
	}
	for (int iii=0;iii<NPAR;iii++) temppar[iii]=PR[iii];
	if (checkpar()) {F1=NEGLI_LH;} else {
        F1 = FUN(temppar,tempnpar);}             
i101:	for (J=0;J<N;J++) {
        if (ICP[J]==0) goto i103;
        if (KK==1)ST[J]=ST[J]*0.5;
i20:      K=2; 
        PR[J]=PR[J]-ST[J];
i102:  	for (int iii=0;iii<NPAR;iii++) temppar[iii]=PR[iii];
	if (checkpar()) {F=NEGLI_LH;} else {
        F = FUN(temppar,tempnpar);}             
        NFUN=NFUN+1; 
	if(F1>=F) {goto i22;} else {goto i23;}
i22:      K=K+1; 
        PR[J]=PR[J]+2.*ST[J]; 
	if(K==3) {goto i31;}else{goto i103;}
i31:      ST[J]=-ST[J]*.5; 
        goto i102; 
i23:      ST[J]=ST[J]*1.5; 
        F1=F;
i103:;     
	}
	for (J=0;J<N;J++) {
        if (ICP[J]==0) goto i24; 
        X=2.*PR[J]-Y[J];
        Z=PR[J]-Y[J];
	if (Z>0.) {goto i202;} 
	if (Z<0.) {goto i201;} else {goto i203;} 
i201:    if ((GL[J]-X)>=0.) {goto i211;} else {goto i203;}
i211:     X=GL[J]+eps;
        goto i203;
i202:    if ((GR[J]-X)<=0.) {goto i212;} else {goto i203;} 
i212:     X=GR[J]-eps;
i203:     Y[J]=PR[J]; 
        PR[J]=X;
i24:;
	}
        K=5; 
i104:  	for (int iii=0;iii<NPAR;iii++) temppar[iii]=PR[iii];
	if (checkpar()) {F=NEGLI_LH;} else {
        F = FUN(temppar,tempnpar);}             
        NFUN=NFUN+1; 
	if ((F-F1)<=0.) {goto i25;} else {goto i50;}
i25:      K=K+1; 
	for (J=0;J<N;J++) {
        if (ICP[J]==0) goto i26;
        PR[J]=Y[J]*(1.-R)+PR[J]*R; 
i26:;      
	}
	if ((K-15)>=0) {goto i27;} else {goto i104;}
i27:      KK=1;
	for (int iii=0;iii<NPAR;iii++) temppar[iii]=PR[iii];
	if (checkpar()) {F=NEGLI_LH;} else {
        F = FUN(temppar,tempnpar);}             
        NFUN=NFUN+1; 
        goto i53;
i50:      KK=0;
	if ((F-BEGIN-EP)>=0) {goto i53;} else {goto i521;}
i521:;     
	//if (ip.ne.0)write(3,595)
	H=F;
	if(F > NEGLI_LH)goto i599;
	H=F1;
	for (int i=0;i<N;i++) {
i600:     PR[i]=Y[i];
	}
i599:     NITER=NITER+1; 
        goto i512; 
i595:     ;//FORMAT (1X,'putative maximim found')
i53:      KP=0;
	for (J=0;J<N;J++) {
        if (ICP[J] == 0) goto i54; 
        if ((tiabs(PER*PR[J])-tiabs(ST[J]))>=0.) {goto i531;} else {goto i45;}
i45:      if ((tiabs(ST[J])-1.E-06)>=0.) {goto i54;} else {goto i531;}
i531:     KP=KP+1; 
i54:     ;// CONTINUE
	}
        if ((KP-N1)>=0) {goto i521;} else {goto i52;}
i52:      NITER=NITER+1; 
	  if(outmode>1)printout(F);
        if ((NITER-NI)<=0) {goto i51;} else {goto i511;}
i511:     ;//CONTINUE
	//if(IP.ne.0)	write (3,594)
	if(outmode>0) printout(H);
	write(fnam,0);
	return "   /maximum not found/";
i512:     ;//CONTINUE
	//if (ip.eq.0)return
	//call WWOUT(N)
	if(outmode>0) printout(H);
	write(fnam,1);
        return "   /putative maximum found/"; //RETURN
i594:     ;//FORMAT (1X,'maximum not found')
i51:	FF=F1;
	if (F > FF) FF=F;
        //IF (IP-1)55,55,501
i501:	FF=F1;
	if (F > FF) FF=F;
	//call WWOUT(N)
i55:      if ((F-(NEGLI_LH*0.1))<=0.) {goto i551;} else {goto i550;}
i550:     F1=F;
i551:     BEGIN=F1;
        goto i101; 
}

int methgi::checkpar() {
	for (int i=0;i<NPAR;i++) {
	  if (IVEC[i]!=(i+1)) PR[i]=PR[IVEC[i]-1];
	}
	for (int i=0;i<NPAR;i++) {
		if (GL[i]>=GR[i]) return 1;
		if (ST[i]>=(GR[i]-GL[i])) return 1;
		if (PR[i]>GR[i] || PR[i]<GL[i]) return 1;
	}
	return 0;
}

void methgi::write(char *fnout,int ifmax) {
	FILE *OUF;
	OUF=fopen(fnout,"w");
	if (!OUF) ABTMsg("       \n/cannot open output file/ ...",fnout);
	fprintf(OUF,"----------------------------------------------------------------\n");
	if (ifmax) {
        fprintf(OUF,"   /putative maximum found/\n");
	} else {
	fprintf(OUF,"   /maximum not found/\n");
	}
	fprintf(OUF,"----------------------------------------------------------------\n");
	fprintf(OUF,".     /maximal function value/ = %f\n",H);
	fprintf(OUF,"              /number of iterations/ = %d\n",NITER);
	fprintf(OUF,"    /number of function calls/ = %d\n",NFUN);
	fprintf(OUF,"            /number of parameters/ = %d\n",NPAR);
	fprintf(OUF,"----------------------------------------------------------------\n");
	fprintf(OUF," N IVEC ICP             PR  ST  GL  GR\n");
	fprintf(OUF,"----------------------------------------------------------------\n");
	for (int i=0;i<NPAR;i++)
	fprintf(OUF,"%2d%5d%4d %14.6E %f %f %f\n",(i+1),IVEC[i],ICP[i],PR[i],ST[i],GL[i],GR[i]);
	fprintf(OUF,"----------------------------------------------------------------\n");
	fclose(OUF);
}

void methgi::printout(double curlh) {
	printf("------------------------------------------------------\n");
	printf("  /function value/ = %f\n",curlh);
	printf("              /iteration/ = %d\n",NITER);
	printf("   /function calls/ = %d\n",NFUN);
	printf("------------------------------------------------------\n");
	printf(" N IVEC ICP             PR  ST  GL  GR\n");
	printf("------------------------------------------------------\n");
	for (int i=0;i<NPAR;i++)
	printf("%2d%5d%4d %14.6E %f %f %f\n",(i+1),IVEC[i],ICP[i],PR[i],ST[i],GL[i],GR[i]);
	printf("------------------------------------------------------\n");
}


void methgi::printout() {
	printf("------------------------------------------------------\n");
	printf(" N IVEC ICP             PR  ST  GL  GR\n");
	printf("------------------------------------------------------\n");
	for (int i=0;i<NPAR;i++)
	printf("%2d%5d%4d %14.6E %f %f %f\n",(i+1),IVEC[i],ICP[i],PR[i],ST[i],GL[i],GR[i]);
	printf("------------------------------------------------------\n");
}

double methgi::getLH() {
	return H;
}

double methgi::value(int pnum) {
        if (pnum>=NPAR) ABTMsg("  \n/parameter does not exist/",pnum); 
        if (pnum<0) ABTMsg("  \n/parameter does not exist/",pnum); 
        return PR[pnum];
}

double methgi::left95(LHfunction FUN,int numpar) {
	double LH0 = FUN(PR,NPAR);
	double LH1 = 0.0;
	double delta = 1.0;
	double prsave = PR[numpar];
	double prR = prsave;
	double prL = GL[numpar];
	double prc = (prR+prL)/2.0;

	while (tiabs(delta)>1.e-6 & tiabs(GL[numpar]-prc)>1.e-8 & tiabs(GR[numpar]-prc)>1.e-8) {
		PR[numpar] = prc;
		LH1 = FUN(PR,NPAR);
		delta = 2.0 * (LH0 - LH1) - 3.841455;
//		cout<<"lh0="<<LH0<<" lh1="<<LH1<<" prc="<<prc<<" "<<prL<<" "<<prR<<" "<<delta<<"\n";
		if (delta>0) {
			prL = prc;
			prc = (prR + prL)/2.0;
		} else {
			prR = prc;
			prc = (prR + prL)/2.0;
		}

	}
	PR[numpar]=prsave;
	return prc;
}
double methgi::right95(LHfunction FUN,int numpar) {
	double LH0 = FUN(PR,NPAR);
	double LH1 = 0.0;
	double delta = 1.0;
	double prsave = PR[numpar];
	double prL = prsave;
	double prR = GR[numpar];
	double prc = (prR+prL)/2.0;

	while (tiabs(delta)>1.e-6 & tiabs(GL[numpar]-prc)>1.e-8 & tiabs(GR[numpar]-prc)>1.e-8) {
		PR[numpar] = prc;
		LH1 = FUN(PR,NPAR);
		delta = 2.0 * (LH0 - LH1) - 3.841455;
//		cout<<"lh0="<<LH0<<" lh1="<<LH1<<" prc="<<prc<<" "<<prL<<" "<<prR<<" "<<delta<<"\n";
		if (delta<0) {
			prL = prc;
			prc = (prR + prL)/2.0;
		} else {
			prR = prc;
			prc = (prR + prL)/2.0;
		}

	}
	PR[numpar]=prsave;
	return prc;
}

int methgi::npar() {
	return NPAR;
}

double *methgi::parlist() {
	for(int i=0;i<NPAR;i++) list[i]=PR[i];
	return list;
}

void ABTMsg(char *msg) {
	cout << msg << endl;
	exit(1);
}

void ABTMsg(char *msg1, char *msg2) {
	cout << msg1 << " " << msg2 << endl;
	exit(1);
}

void ABTMsg(char *msg1, int msg2) {
	cout << msg1 << " " << msg2 << endl;
	exit(1);
}

void ABTMsg(char *msg1, double msg2) {
	cout << msg1 << " " << msg2 << endl;
	exit(1);
}

double tiabs(double v) {
	return v > 0 ? v : -v;
}
